﻿using Timer = System.Timers.Timer;

namespace Model
{
    /// <summary>
    /// Defines a bomb that can be exploded, and can explode on its own. It sends an event when the explosion happends.
    /// </summary>
    internal class Bomb : IDisposable
    {
        #region Fields

        private Timer _timer;

        #endregion

        #region Properties

        /// <summary>
        /// The X position of the bomb.
        /// </summary>
        public int X { get; }

        /// <summary>
        /// The Yposition of the bomb.
        /// </summary>
        public int Y { get; }

        /// <summary>
        /// True, if the bomb does not explode on its own.
        /// </summary>
        public bool IsDetonable { get; }

        #endregion

        #region Events

        /// <summary>
        /// Occurs when the bomb explodes.
        /// </summary>
        public event EventHandler? BombIsExploding;

        #endregion

        #region Constructors

        /// <summary>
        /// Initializes an instance of the Model.Bomb class.
        /// </summary>
        /// <param name="x">The X position of the bomb.</param>
        /// <param name="y">The Y position of the bomb.</param>
        /// <param name="isDetonable">True, if the bomb should not explode on its own, false, if it should explode after a set amount of time.</param>
        public Bomb(int x, int y, bool isDetonable)
        {
            X = x;
            Y = y;
            IsDetonable = isDetonable;

            _timer = new Timer(5000); // adjust later for better gameplay
            _timer.AutoReset = false;
            _timer.Elapsed += (sender, e) => BombIsExploding?.Invoke(this, EventArgs.Empty);

            if (!isDetonable)
            {
                _timer.Start();
            }
        }

        /// \cond DOXYGEN_SHOULD_SKIP_THIS
        /// <summary>
        /// Only for testing! Initializes a new instance of the Model.Bomb class for testing.
        /// </summary>
        internal Bomb(int x, int y, bool isDetonable, int interval) : this(x, y, isDetonable)
        {
            _timer.Interval = interval;
        }
        /// \endcond

        #endregion

        #region Public Methods

        /// <summary>
        /// Explodes the bomb, stopping it's timer if it had one.
        /// </summary>
        public void ExplodeNow()
        {
            _timer.Stop();
            BombIsExploding?.Invoke(this, EventArgs.Empty);
        }

        #endregion

        #region IDisposable

        private bool _disposed = false;

        ~Bomb() => Dispose(false);

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
                    _timer.Dispose();
                }

                _disposed = true;
            }
        }

        #endregion
    }
}
