﻿using Model.Utils;

using static Model.Utils.LowerFieldType;
using static Model.Utils.UpperFieldType;

namespace Model
{
    /// <summary>
    /// The ClumsyMonster is a subclass of the Monster class. It moves like the SmartMonster, but it has a 30% chance that the monster moves to the wrong direction.
    /// </summary>
    internal class ClumsyMonster : Monster
    {
        /// <summary>
        /// Initializes a new instance of the ClumsyMonster class.
        /// </summary>
        /// <param name="x">Initial X position.</param>
        /// <param name="y">Initial Y position.</param>
        /// <param name="indexer">The indexer function of the map.</param>
        public ClumsyMonster(int x, int y, Func<int, int, (LowerFieldType, UpperFieldType)> indexer) : base(x, y, indexer)
        {
            _moveTimer.Interval = 1500; // same speed as the BasicMonster
            _moveTimer.Start();
            _canStepOnTheseLowers = [LowerFieldType.Empty, MoreBombsEffect, BiggerBombsEffect, DetonatorEffect, SkatesEffect, InvincibilityEffect, GhostEffect, BarrierEffect, Debuff];
            _canStepOnTheseUppers = [UpperFieldType.Empty, BluePlayer, RedPlayer];
        }

        /// \cond DOXYGEN_SHOULD_SKIP_THIS
        /// <summary>
        /// Only for testing! Initializes a new instance of the Model.ClumsyMonster class for testing.
        /// </summary>
        internal ClumsyMonster(int x, int y, Func<int, int, (LowerFieldType, UpperFieldType)> indexer, int interval) : this(x, y, indexer)
        {
            _moveTimer.Interval = interval;
        }
        /// \endcond

        protected override Direction? DoDecideDirectionToMove()
        {
            int fromX = X, fromY = Y;
            Direction direction = _direction;

            List<Direction> valids = GetValidDirections();

            if (valids.Count == 0)
            {
                return null;
            }
            if (valids.Count == 1)
            {
                direction = valids[0];
            }
            else
            {
                // determine if we need to change direction
                LowerFieldType lower;
                UpperFieldType upper;
                switch (direction)
                {
                    case Direction.Up:
                        (lower, upper) = _mapIndexer(X, Y - 1);
                        break;
                    case Direction.Down:
                        (lower, upper) = _mapIndexer(X, Y + 1);
                        break;
                    case Direction.Left:
                        (lower, upper) = _mapIndexer(X - 1, Y);
                        break;
                    case Direction.Right:
                        (lower, upper) = _mapIndexer(X + 1, Y);
                        break;
                    default:
                        throw new ArgumentOutOfRangeException($"Error with the direction: {direction}");
                }

                // if we need to change direction, what is the best direction to go
                // if we don't need to change direction, we will keep the current direction
                if (!_canStepOnTheseLowers.Contains(lower) || !_canStepOnTheseUppers.Contains(upper))
                {
                    Direction? newDirection = DirectionToClosestEnemy();
                    if (newDirection != null)
                    {
                        // there is a 70% chance that the monster will make a decision based on the closest player
                        if (_random.Next(10) < 3)
                        {
                            valids.Remove(direction);
                            direction = valids[_random.Next(valids.Count)];
                        }
                        else
                        {
                            direction = (Direction)newDirection;
                        }
                    }
                    else
                    {
                        // can't find any player, random movement
                        direction = valids[_random.Next(valids.Count)];
                    }
                }
            }

            return direction;
        }

        #region IDisposable

        private bool _disposed = false;

        ~ClumsyMonster() => Dispose(false);

        protected override void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {

                }

                _disposed = true;

                base.Dispose(disposing);
            }
        }

        #endregion
    }
}
