﻿using Model.Utils;

using Timer = System.Timers.Timer;

namespace Model
{
    /// <summary>
    /// Simulates a game of Bomberman. Handles the inputs of the players, start and end of a competition, and supplies several events which signal changes in the gamestate.
    /// </summary>
    public partial class GameModel : IDisposable
    {
        #region Fields

        private static readonly LowerFieldType[] allEffects = [LowerFieldType.Debuff, LowerFieldType.MoreBombsEffect, LowerFieldType.BiggerBombsEffect, LowerFieldType.DetonatorEffect, LowerFieldType.SkatesEffect, LowerFieldType.BarrierEffect, LowerFieldType.InvincibilityEffect, LowerFieldType.GhostEffect];

        private readonly Func<int, int, (LowerFieldType, UpperFieldType)> _indexerFunction;

        private bool _wasTerminated = false;
        private int _layoutNumber;
        private int _pointsToWin;
        private int _bluePoints;
        private int _redPoints;
        private readonly int _battleRoyaleMaxTime = 15;
        private readonly int _battleRoyaleReduce = 1;
        private readonly int _battleRoyaleMinTime = 5;
        private int _battleRoyaleStep;
        private int _battleRoyaleTimeLeft;

        private LowerFieldType[,] _lowerLayer;
        private UpperFieldType[,] _upperLayer;
        private int[,] _bombLevel;

        private Player _bluePlayer = null!;
        private Player _redPlayer = null!;
        private List<Bomb> _blueBombs;
        private List<Bomb> _redBombs;
        private List<Monster> _monsters;

        private Timer _battleRoyaleTimer;
        private Timer _onePlayerDeadTimer;
        private Random _random = new Random();

        #endregion

        #region Properties

        /// <summary>
        /// The point count of the blue player.
        /// </summary>
        public int BluePoints
        {
            get => _bluePoints;
            private set => _bluePoints = value;
        }

        /// <summary>
        /// The point count of the red player.
        /// </summary>
        public int RedPoints
        {
            get => _redPoints;
            private set => _redPoints = value;
        }

        #endregion

        #region Events

        /// <summary>Occurs when a new round starts.</summary>
        public event EventHandler<NewRoundEventArgs>? NewRound;
        /// <summary>Occurs when a player or monster moves from a tile to another.</summary>
        public event EventHandler<MoveEventArgs>? EntityMoved;
        /// <summary>Occurs when a player or monster dies.</summary>
        public event EventHandler<PositionEventArgs>? EntityDies;
        /// <summary>Occurs when a 'lower layer' field gets changed.</summary>
        public event EventHandler<FieldChangedEventArgs>? FieldChanged;
        /// <summary>Occurs when a bomb explosion happens on a tile.</summary>
        public event EventHandler<PositionEventArgs>? BombExplosion;
        /// <summary>Occurs when the competition is over and a player wins.</summary>
        public event EventHandler<PlayerEventArgs>? CompetitionOver;
        /// <summary>Occurs when the time remaining in a battle royale loop changed.</summary>
        public event EventHandler<IntEventArgs>? BattleRoyaleTimeChanged;
        /// <summary>Occurs when one of the player's special power statuses get updated.</summary>
        public event EventHandler<PlayerStatusEventArgs>? PlayerStatusChanged;

        #endregion

        #region Constructors

        /// <summary>
        /// Initializes a new instance of the Model.GameModel class.
        /// </summary>
        public GameModel()
        {
            _lowerLayer = new LowerFieldType[HEIGHT, WIDTH];
            _upperLayer = new UpperFieldType[HEIGHT, WIDTH];
            _blueBombs = new List<Bomb>();
            _redBombs = new List<Bomb>();
            _monsters = new List<Monster>();
            _indexerFunction = Indexer;
            _bombLevel = new int[HEIGHT, WIDTH];
            _onePlayerDeadTimer = new Timer(5000);
            _onePlayerDeadTimer.AutoReset = false;
            _onePlayerDeadTimer.Elapsed += RoundOver;

            _battleRoyaleTimer = new Timer(1000);
            _battleRoyaleTimer.AutoReset = true;
            _battleRoyaleTimer.Elapsed += BattleRoyaleSecondPassed;
        }

        /// \cond DOXYGEN_SHOULD_SKIP_THIS
        /// <summary>
        /// Only for testing! Initializes a new instance of the Model.GameModel class for testing.
        /// </summary>
        internal GameModel(LowerFieldType[,] lowerLayer, UpperFieldType[,] upperLayer, int layoutNumber, int battleRoyaleTime) : this()
        {
            _layoutNumber = layoutNumber;
            if (lowerLayer.Length != upperLayer.Length)
            {
                throw new ArgumentException("Layers were not compatible (different sizes)!");
            }

            if (layoutNumber == 1)
            {
                LayoutOneLowerLayer = lowerLayer;
                LayoutOneUpperLayer = upperLayer;
            }
            else if (layoutNumber == 2)
            {
                LayoutTwoLowerLayer = lowerLayer;
                LayoutTwoUpperLayer = upperLayer;
            }
            else if (layoutNumber == 3)
            {
                LayoutThreeLowerLayer = lowerLayer;
                LayoutThreeUpperLayer = upperLayer;
            }

            HEIGHT = lowerLayer.GetLength(0);
            WIDTH = lowerLayer.GetLength(1);
            _lowerLayer = new LowerFieldType[HEIGHT, WIDTH];
            _upperLayer = new UpperFieldType[HEIGHT, WIDTH];
            _battleRoyaleMaxTime = battleRoyaleTime;
        }
        /// \endcond

        #endregion

        #region Public Methods

        /// <summary>
        /// Starts a new competition.
        /// </summary>
        /// <param name="points">The amount of points needed to win.</param>
        /// <param name="layoutNum">The number of the layout (1, 2, 3).</param>
        public void NewCompetitionStarts(int points, int layoutNum)
        {
            _pointsToWin = points;
            _layoutNumber = layoutNum;
            _wasTerminated = false;

            _bluePoints = 0;
            _redPoints = 0;

            StartNewRound();
        }

        /// <summary>
        /// Moves the player in the direction, if it can move and it is a valid move.
        /// </summary>
        /// <param name="player">The player trying to move.</param>
        /// <param name="direction">The direction in which the player is trying to move.</param>
        public void Move(PlayerType player, Direction direction)
        {
            if (player == PlayerType.Blue)
            {
                MovePlayer(_bluePlayer, direction);
            }
            else
            {
                MovePlayer(_redPlayer, direction);
            }
        }

        /// <summary>
        /// The given player places a bomb if it has any and there is room for it, or detonates their detonatable bombs if it can place no more.
        /// </summary>
        /// <param name="player">The player trying to place a bomb.</param>
        public void PlaceBomb(PlayerType player)
        {
            Player bomber = player == PlayerType.Blue ? _bluePlayer : _redPlayer;
            List<Bomb> bomblist = player == PlayerType.Blue ? _blueBombs : _redBombs;

            if (bomber.IsDead)
                return;

            bool placebomb = false, detonatable = false;

            if (bomber.HasDetonator)
            {
                // bomber has detonator
                if (bomber.MaxBombCount == bomblist.Count)
                {

                    // we're exploding the detonatable bombs
                    for (int i = bomblist.Count - 1; i >= 0; --i)
                    {
                        if (bomblist[i].IsDetonable)
                            bomblist[i].ExplodeNow();
                    }
                    // This explodes the bombs, which invokes the Bomb_BombIsExploding method, which removes
                    // the bomb from this list. Meaning, that if we use a foreach, the iterator gets invalidated,
                    // and an exception gets thrown in the foreach. If we index upwards, and one gets removed,
                    // we will "skip" a bomb. Hence, we index downwards.
                }
                else
                {
                    // we're placing a detonatable bomb
                    if (!bomber.CantPlaceBombs &&
                        _lowerLayer[bomber.Y, bomber.X] == LowerFieldType.Empty)
                    {
                        placebomb = true;
                        detonatable = true;
                    }
                }
            }
            else
            {
                // bomber has no detonator
                if (!bomber.CantPlaceBombs &&
                    bomber.MaxBombCount != bomblist.Count &&
                    _lowerLayer[bomber.Y, bomber.X] == LowerFieldType.Empty)
                {
                    placebomb = true;
                }

            }

            if (placebomb)
            {
                Bomb bomb = new(bomber.X, bomber.Y, detonatable);
                bomb.BombIsExploding += Bomb_BombIsExploding;
                bomblist.Add(bomb);
                _lowerLayer[bomber.Y, bomber.X] = LowerFieldType.Bomb;
                InvokeFieldChanged(bomber.X, bomber.Y, LowerFieldType.Bomb);
            }
        }

        /// <summary>
        /// The given player places a barrier if it has any and there is room for it.
        /// </summary>
        /// <param name="p">The player trying to place a barrier.</param>
        public void PlaceBarrier(PlayerType p)
        {
            Player player = p == PlayerType.Blue ? _bluePlayer : _redPlayer;
            if (player.IsDead || !player.CanPlaceBarrier)
            {
                return;
            }

            if (_lowerLayer[player.Y, player.X] == LowerFieldType.Empty)
            {
                player.BarrierPlaced();
                _lowerLayer[player.Y, player.X] = p == PlayerType.Blue ? LowerFieldType.BlueBarrier : LowerFieldType.RedBarrier;
                InvokeFieldChanged(player.X, player.Y, p == PlayerType.Blue ? LowerFieldType.BlueBarrier : LowerFieldType.RedBarrier);
            }

        }

        /// <summary>
        /// Terminates the running competition.
        /// </summary>
        public void TerminateCompetition()
        {
            _wasTerminated = true;
            _battleRoyaleTimer.Stop();
            _onePlayerDeadTimer.Stop();

            _bluePlayer.PlayerDied -= PlayerDied;
            _bluePlayer.StatusChanged -= Player_StatusChanged;
            _bluePlayer.InvincibilityBuffIsOver -= Player_InvincibilityOver;
            _bluePlayer.GhostBuffIsOver -= Player_GhostOver;

            _redPlayer.PlayerDied -= PlayerDied;
            _redPlayer.StatusChanged -= Player_StatusChanged;
            _redPlayer.InvincibilityBuffIsOver -= Player_InvincibilityOver;
            _redPlayer.GhostBuffIsOver -= Player_GhostOver;

            _bluePlayer.PlayerDies();
            _redPlayer.PlayerDies();

            foreach (Bomb bomb in _blueBombs)
            {
                bomb.BombIsExploding -= Bomb_BombIsExploding;
                bomb.ExplodeNow();
            }
            foreach (Bomb bomb in _redBombs)
            {
                bomb.BombIsExploding -= Bomb_BombIsExploding;
                bomb.ExplodeNow();
            }

            foreach (Monster monster in _monsters)
            {
                monster.MonsterWantsToMove -= Monster_MonsterWantsToMove;
                monster.MonsterDied();
            }
        }

        #endregion

        #region Private Methods

        private void MovePlayer(Player player, Direction direction)
        {
            int fromX = player.X;
            int fromY = player.Y;
            bool success = player.PlayerMoves(direction);
            if (success)
            {
                if (player.IsDead)
                {
                    EntityDies?.Invoke(this, new PositionEventArgs(fromX, fromY));
                    _upperLayer[fromY, fromX] = UpperFieldType.Empty;
                }
                else
                {
                    int toX = player.X;
                    int toY = player.Y;
                    EntityMoved?.Invoke(this, new MoveEventArgs(fromX, fromY, toX, toY, direction));
                    _upperLayer[toY, toX] = _upperLayer[fromY, fromX];
                    _upperLayer[fromY, fromX] = UpperFieldType.Empty;
                    if (allEffects.Contains(_lowerLayer[toY, toX]))
                    {
                        _lowerLayer[toY, toX] = LowerFieldType.Empty;
                        InvokeFieldChanged(toX, toY, LowerFieldType.Empty);
                    }
                    if (player.InstantBombs)
                    {
                        if (_lowerLayer[toY, toX] == LowerFieldType.Empty)
                        {
                            if (player == _redPlayer)
                            {
                                PlaceInstantBomb(PlayerType.Red);
                            }
                            else
                            {
                                PlaceInstantBomb(PlayerType.Blue);
                            }

                        }
                    }
                }
            }
        }

        private void StartNewRound()
        {
            _bluePlayer?.Dispose();
            _redPlayer?.Dispose();
            foreach (Bomb bomb in _blueBombs)
                bomb.Dispose();
            foreach (Bomb bomb in _redBombs)
                bomb.Dispose();
            foreach (Monster monster in _monsters)
                monster.Dispose();

            _blueBombs.Clear();
            _redBombs.Clear();
            _monsters.Clear();

            LowerFieldType[,] lowerToCopyFrom = LayoutOneLowerLayer;
            UpperFieldType[,] upperToCopyFrom = LayoutOneUpperLayer;

            if (_layoutNumber == 1)
            {
                lowerToCopyFrom = LayoutOneLowerLayer;
                upperToCopyFrom = LayoutOneUpperLayer;
            }
            else if (_layoutNumber == 2)
            {
                lowerToCopyFrom = LayoutTwoLowerLayer;
                upperToCopyFrom = LayoutTwoUpperLayer;
            }
            else
            {
                lowerToCopyFrom = LayoutThreeLowerLayer;
                upperToCopyFrom = LayoutThreeUpperLayer;
            }

            int blueX = -1;
            int blueY = -1;
            int redX  = -1;
            int redY  = -1;

            for (int x = 0; x < WIDTH; x++)
            {
                for (int y = 0; y < HEIGHT; y++)
                {
                    _lowerLayer[y, x] = lowerToCopyFrom[y, x];
                    _upperLayer[y, x] = upperToCopyFrom[y, x];
                    if (_upperLayer[y, x] == UpperFieldType.BasicMonster)
                    {
                        (UpperFieldType monsterType, Monster monster) = Monster.RandomMonsterFactory(x, y, _indexerFunction);
                        _upperLayer[y, x] = monsterType;
                        _monsters.Add(monster);
                        monster.MonsterWantsToMove += Monster_MonsterWantsToMove;
                    }

                    _bombLevel[y, x] = 0;

                    if (_upperLayer[y, x] == UpperFieldType.BluePlayer)
                    {
                        blueX = x;
                        blueY = y;
                    }
                    if (_upperLayer[y, x] == UpperFieldType.RedPlayer)
                    {
                        redX = x;
                        redY = y;
                    }
                }
            }

            if (blueX == -1 || blueY == -1 || redX == -1 || redY == -1)
                throw new Exception("The layout is missing a player!");

            _bluePlayer = new Player(blueX, blueY, _indexerFunction);
            _redPlayer = new Player(redX, redY, _indexerFunction);
            _bluePlayer.PlayerDied += PlayerDied;
            _redPlayer.PlayerDied += PlayerDied;
            _bluePlayer.InvincibilityBuffIsOver += Player_InvincibilityOver;
            _redPlayer.InvincibilityBuffIsOver += Player_InvincibilityOver;
            _bluePlayer.GhostBuffIsOver += Player_GhostOver;
            _redPlayer.GhostBuffIsOver += Player_GhostOver;
            _bluePlayer.StatusChanged += Player_StatusChanged;
            _redPlayer.StatusChanged += Player_StatusChanged;

            _battleRoyaleTimeLeft = _battleRoyaleMaxTime;
            _battleRoyaleStep = 0;
            BattleRoyaleTimeChanged?.Invoke(this, new IntEventArgs(_battleRoyaleTimeLeft));
            _battleRoyaleTimer.Start();

            NewRound?.Invoke(this, new NewRoundEventArgs(_lowerLayer, _upperLayer));
        }

        private (LowerFieldType, UpperFieldType) Indexer(int x, int y)
        {
            return (_lowerLayer[y, x], _upperLayer[y, x]);
        }

        private void MaybeDestroyEntityOnTile(int x, int y, bool destroyInvinciblePlayers)
        {
            switch (_upperLayer[y, x])
            {
                case UpperFieldType.BluePlayer:
                    if (!_bluePlayer.IsInvincible || destroyInvinciblePlayers)
                    {
                        _upperLayer[y, x] = UpperFieldType.Empty;
                        InvokeEntityDies(x, y);
                        _bluePlayer.PlayerDies();
                    }
                    return;
                case UpperFieldType.RedPlayer:
                    if (!_redPlayer.IsInvincible || destroyInvinciblePlayers)
                    {
                        _upperLayer[y, x] = UpperFieldType.Empty;
                        InvokeEntityDies(x, y);
                        _redPlayer.PlayerDies();
                    }
                    return;
                case UpperFieldType.BasicMonster:
                case UpperFieldType.GhostMonster:
                case UpperFieldType.SmartMonster:
                case UpperFieldType.ClumsyMonster:
                    foreach (Monster monster in _monsters)
                    {
                        if (monster.X == x && monster.Y == y)
                        {
                            monster.MonsterDied();
                            monster.Dispose();
                            _upperLayer[y, x] = UpperFieldType.Empty;
                            InvokeEntityDies(x, y);
                            return;
                        }
                    }
                    throw new InvalidOperationException($"Monster on ({x},{y}) was not found in {nameof(_monsters)}!");
                default:
                    return;
            }
        }

        private async Task ExecuteExplosionOnTile(int x, int y)
        {
            _upperLayer[y, x] = UpperFieldType.Explosion;
            ++_bombLevel[y, x];
            InvokeBombExplosion(x, y);

            await Task.Delay(1000);

            if (--_bombLevel[y, x] == 0)
            {
                _upperLayer[y, x] = UpperFieldType.Empty;
            }
        }

        private LowerFieldType GenerateRandomEffect()
        {
            int num = _random.Next(100);
            if (num < 50)
                return LowerFieldType.Empty;

            int max = allEffects.Length;
            num = _random.Next(0, max);
            return allEffects[num];
        }

        private void PlaceInstantBomb(PlayerType player)
        {
            Player bomber = player == PlayerType.Blue ? _bluePlayer : _redPlayer;
            List<Bomb> bomblist = player == PlayerType.Blue ? _blueBombs : _redBombs;

            if (bomber.MaxBombCount == bomblist.Count)
                return;

            PlaceBomb(player);
        }

        #endregion

        #region Private Event Methods

        private void BattleRoyaleSecondPassed(object? sender, EventArgs e)
        {
            _battleRoyaleTimeLeft--;
            BattleRoyaleTimeChanged?.Invoke(this, new IntEventArgs(_battleRoyaleTimeLeft));

            if (_battleRoyaleTimeLeft == 0)
            {

                HashSet<(int, int)> putOuterWalls = new HashSet<(int, int)>();
                for (int x = 0; x < WIDTH; x++)
                {
                    for (int y = 0; y < HEIGHT; y++)
                    {
                        if (_lowerLayer[y, x] == LowerFieldType.OuterWall)
                        {
                            if (y - 1 >= 0)
                                putOuterWalls.Add((x, y - 1));
                            if (y + 1 < HEIGHT)
                                putOuterWalls.Add((x, y + 1));
                            if (x - 1 >= 0)
                                putOuterWalls.Add((x - 1, y));
                            if (x + 1 < WIDTH)
                                putOuterWalls.Add((x + 1, y));
                        }
                    }
                }
                foreach ((int x, int y) in putOuterWalls)
                {
                    MaybeDestroyEntityOnTile(x, y, true);
                    switch (_lowerLayer[y, x])
                    {
                        case LowerFieldType.InnerWall:
                        case LowerFieldType.Empty:
                        case LowerFieldType.Box:
                        case LowerFieldType.MoreBombsEffect:
                        case LowerFieldType.BiggerBombsEffect:
                        case LowerFieldType.DetonatorEffect:
                        case LowerFieldType.SkatesEffect:
                        case LowerFieldType.InvincibilityEffect:
                        case LowerFieldType.GhostEffect:
                        case LowerFieldType.BarrierEffect:
                        case LowerFieldType.Debuff:
                            _lowerLayer[y, x] = LowerFieldType.OuterWall;
                            InvokeFieldChanged(x, y, LowerFieldType.OuterWall);
                            break;
                        case LowerFieldType.BlueBarrier:
                            _lowerLayer[y, x] = LowerFieldType.OuterWall;
                            InvokeFieldChanged(x, y, LowerFieldType.OuterWall);
                            _bluePlayer.BarrierDestroyed();
                            break;
                        case LowerFieldType.RedBarrier:
                            _lowerLayer[y, x] = LowerFieldType.OuterWall;
                            InvokeFieldChanged(x, y, LowerFieldType.OuterWall);
                            _redPlayer.BarrierDestroyed();
                            break;
                        case LowerFieldType.Bomb:
                            foreach (Bomb bomb in _blueBombs)
                            {
                                if (bomb.X == x && bomb.Y == y)
                                {
                                    bomb.BombIsExploding -= Bomb_BombIsExploding;
                                    _blueBombs.Remove(bomb);
                                    bomb.Dispose();
                                    _lowerLayer[y, x] = LowerFieldType.OuterWall;
                                    InvokeFieldChanged(x, y, LowerFieldType.OuterWall);
                                    break;
                                }
                            }
                            foreach (Bomb bomb in _redBombs)
                            {
                                if (bomb.X == x && bomb.Y == y)
                                {
                                    bomb.BombIsExploding -= Bomb_BombIsExploding;
                                    _redBombs.Remove(bomb);
                                    bomb.Dispose();
                                    _lowerLayer[y, x] = LowerFieldType.OuterWall;
                                    InvokeFieldChanged(x, y, LowerFieldType.OuterWall);
                                    break;
                                }
                            }
                            break;
                    }
                }

                _battleRoyaleTimeLeft = Math.Max(_battleRoyaleMinTime, _battleRoyaleMaxTime - (++_battleRoyaleStep * _battleRoyaleReduce));
                BattleRoyaleTimeChanged?.Invoke(this, new IntEventArgs(_battleRoyaleTimeLeft));
            }
        }

        private void Player_StatusChanged(object? sender, StatusEventArgs e)
        {
            if (sender is Player player)
                PlayerStatusChanged?.Invoke(this, new(e.Power, e.Show, player == _bluePlayer ? PlayerType.Blue : PlayerType.Red));
            else
                throw new ArgumentException("No player invoked the event!");
        }

        private void Monster_MonsterWantsToMove(object? sender, MoveEventArgs e)
        {
            if (sender is Monster monster)
            {
                bool success;
                lock (_upperLayer)
                {
                    if (_upperLayer[e.ToY, e.ToX] is UpperFieldType.Empty or UpperFieldType.Explosion or
                                                     UpperFieldType.BluePlayer or UpperFieldType.RedPlayer &&
                        _lowerLayer[e.ToY, e.ToX] != LowerFieldType.OuterWall &&
                        (monster is GhostMonster || _lowerLayer[e.ToY, e.ToX] is not LowerFieldType.BlueBarrier or LowerFieldType.RedBarrier))
                    {
                        success = true;
                        if (_upperLayer[e.ToY, e.ToX] == UpperFieldType.BluePlayer)
                        {
                            if (_bluePlayer.IsInvincible) return;
                            _bluePlayer.PlayerDies();
                        }
                        else if (_upperLayer[e.ToY, e.ToX] == UpperFieldType.RedPlayer)
                        {
                            if (_redPlayer.IsInvincible) return;
                            _redPlayer.PlayerDies();
                        }
                        else if (_upperLayer[e.ToY, e.ToX] == UpperFieldType.Explosion)
                        {
                            monster.MonsterDied();
                            monster.Dispose();
                            InvokeEntityDies(e.FromX, e.FromY);
                            return;
                        }
                        _upperLayer[e.ToY, e.ToX] = _upperLayer[e.FromY, e.FromX];
                        _upperLayer[e.FromY, e.FromX] = UpperFieldType.Empty;
                        monster.SuccessfulMove(e.ToX, e.ToY, e.Direction);
                        // this invoke is enough, because for the view and viewmodel it doesn't matter if the player died or just the monster moved to that position
                        EntityMoved?.Invoke(this, new MoveEventArgs(e.FromX, e.FromY, e.ToX, e.ToY, e.Direction));
                    }
                    else
                    {
                        success = false;
                    }
                }

                if (!success)
                {
                    monster.UnsuccessfulMove();
                    return;
                }
            }
            else
                throw new ArgumentException("No monster invoked the event, or method was used improperly!");
        }

        private async void Bomb_BombIsExploding(object? sender, EventArgs e)
        {

            if (sender is Bomb bomb)
            {
                Player bomber = _bluePlayer;
                
                // this assumes that exactly one list contains the exploding bomb
                if (!_blueBombs.Remove(bomb))
                {
                    _redBombs.Remove(bomb);
                    bomber = _redPlayer;
                    if (_redPlayer.InstantBombs)
                    {
                        PlaceInstantBomb(PlayerType.Red);
                    }
                }
                else
                {
                    if (_bluePlayer.InstantBombs)
                    {
                        PlaceInstantBomb(PlayerType.Blue);
                    }
                }

                int x = bomb.X, y = bomb.Y;

                _lowerLayer[y, x] = LowerFieldType.Empty;
                InvokeFieldChanged(x, y, LowerFieldType.Empty);

                MaybeDestroyEntityOnTile(x, y, false);
                if (!(_upperLayer[y, x] is UpperFieldType.BluePlayer && _bluePlayer.IsInvincible ||
                    _upperLayer[y, x] is UpperFieldType.RedPlayer && _redPlayer.IsInvincible))
                    _ = ExecuteExplosionOnTile(x, y);

                List<Direction> dirs = [Direction.Up, Direction.Down, Direction.Right, Direction.Left];
                int maxrange = bomber.HasSmallBombs ? 1 : bomber.MaxBombRange;

                for (int explosiondistance = 1; explosiondistance <= maxrange; ++explosiondistance)
                {
                    await Task.Delay(200);

                    #region Indexing Explanation
                    // This code can remove elements from the dirs list. Meaning, that if we use a foreach,
                    // the iterator gets invalidated, and an exception gets thrown in the foreach. If we index
                    // upwards, and one gets removed, we will "skip" a direction. Hence, we index downwards.
                    #endregion
                    for (int dirindex = dirs.Count - 1; dirindex >= 0; --dirindex)
                    {
                        Direction dir = dirs[dirindex];
                        x = bomb.X;
                        y = bomb.Y;
                        switch (dir)
                        {
                            case Direction.Up:    y -= explosiondistance; break;
                            case Direction.Down:  y += explosiondistance; break;
                            case Direction.Left:  x -= explosiondistance; break;
                            case Direction.Right: x += explosiondistance; break;
                            default:
                                throw new NotSupportedException($"Unknown Direction value received: {dir}");
                        }

                        if (_upperLayer[y, x] is UpperFieldType.BluePlayer && _bluePlayer.IsInvincible ||
                            _upperLayer[y, x] is UpperFieldType.RedPlayer && _redPlayer.IsInvincible)
                        {
                            dirs.Remove(dir);
                        }
                        else
                        {
                            switch (_lowerLayer[y, x])
                            {
                                case LowerFieldType.InnerWall:
                                case LowerFieldType.OuterWall:
                                    dirs.Remove(dir);
                                    break;
                                case LowerFieldType.Box:
                                    dirs.Remove(dir);
                                    _lowerLayer[y, x] = GenerateRandomEffect();
                                    InvokeFieldChanged(x, y, _lowerLayer[y, x]);
                                    MaybeDestroyEntityOnTile(x, y, false);
                                    _ = ExecuteExplosionOnTile(x, y);
                                    break;
                                case LowerFieldType.BlueBarrier:
                                    dirs.Remove(dir);
                                    _lowerLayer[y, x] = LowerFieldType.Empty;
                                    InvokeFieldChanged(x, y, LowerFieldType.Empty);
                                    _bluePlayer.BarrierDestroyed();
                                    MaybeDestroyEntityOnTile(x, y, false);
                                    _ = ExecuteExplosionOnTile(x, y);
                                    break;
                                case LowerFieldType.RedBarrier:
                                    dirs.Remove(dir);
                                    _lowerLayer[y, x] = LowerFieldType.Empty;
                                    InvokeFieldChanged(x, y, LowerFieldType.Empty);
                                    _redPlayer.BarrierDestroyed();
                                    MaybeDestroyEntityOnTile(x, y, false);
                                    _ = ExecuteExplosionOnTile(x, y);
                                    break;
                                case LowerFieldType.Bomb:
                                    dirs.Remove(dir);
                                    foreach (Bomb b in _blueBombs)
                                    {
                                        if (b.X == x && b.Y == y)
                                        {
                                            b.ExplodeNow();
                                            break;
                                        }
                                    }
                                    foreach (Bomb b in _redBombs)
                                    {
                                        if (b.X == x && b.Y == y)
                                        {
                                            b.ExplodeNow();
                                            break;
                                        }
                                    }
                                    break;
                                default:
                                    // empty or effects
                                    MaybeDestroyEntityOnTile(x, y, false);
                                    _ = ExecuteExplosionOnTile(x, y);
                                    break;
                            }
                        }
                    }
                }

                bomb.Dispose();
            }
            else
            {
                throw new ArgumentException("Sender was not an instance of Bomb!", nameof(sender));
            }
        }

        private void PlayerDied(object? sender, EventArgs e)
        {
            if (!_onePlayerDeadTimer.Enabled)
            {
                _onePlayerDeadTimer.Start();
                return;
            }

            _onePlayerDeadTimer.Stop();
            RoundOver(this, EventArgs.Empty);
        }

        private async void RoundOver(object? sender, EventArgs e)
        {
            _battleRoyaleTimer.Stop();

            _bluePlayer.StatusChanged -= Player_StatusChanged;
            _redPlayer.StatusChanged -= Player_StatusChanged;

            foreach (Monster monster in _monsters)
            {
                monster.MonsterWantsToMove -= Monster_MonsterWantsToMove;
                monster.MonsterDied();
            }
            foreach (Bomb bomb in _blueBombs)
            {
                bomb.BombIsExploding -= Bomb_BombIsExploding;
            }
            foreach (Bomb bomb in _redBombs)
            {
                bomb.BombIsExploding -= Bomb_BombIsExploding;
            }

            if (!_bluePlayer.IsDead)
            {
                if (++_bluePoints >= _pointsToWin)
                {
                    _bluePlayer.PlayerDied -= PlayerDied;
                    _bluePlayer.PlayerDies();
                    await Task.Delay(3000);
                    if (!_wasTerminated)
                        CompetitionOver?.Invoke(this, new PlayerEventArgs(PlayerType.Blue));
                    return;
                }
            }
            if (!_redPlayer.IsDead)
            {
                if (++_redPoints >= _pointsToWin)
                {
                    _redPlayer.PlayerDied -= PlayerDied;
                    _redPlayer.PlayerDies();
                    await Task.Delay(3000);
                    if (!_wasTerminated)
                        CompetitionOver?.Invoke(this, new PlayerEventArgs(PlayerType.Red));
                    return;
                }
            }
            _bluePlayer.PlayerDied -= PlayerDied;
            _bluePlayer.PlayerDies();
            _redPlayer.PlayerDied -= PlayerDied;
            _redPlayer.PlayerDies();

            await Task.Delay(3000);
            if (_wasTerminated)
                return;
            StartNewRound();
        }

        private void Player_InvincibilityOver(object? sender, EventArgs e)
        {
            if (sender is Player player)
            {
                if (_bombLevel[player.Y, player.X] > 0)
                {
                    InvokeEntityDies(player.X, player.Y);
                    player.PlayerDies();
                }
            }
        }

        private void Player_GhostOver(object? sender, EventArgs e)
        {
            if (sender is Player player)
            {
                if (!player.JustPlacedBarrier &&
                    _lowerLayer[player.Y, player.X] is LowerFieldType.InnerWall or LowerFieldType.OuterWall or
                    LowerFieldType.Box or LowerFieldType.BlueBarrier or LowerFieldType.RedBarrier)
                {
                    InvokeEntityDies(player.X, player.Y);
                    player.PlayerDies();
                }
            }
        }

        #endregion

        #region Event triggers

        private void InvokeEntityDies(int x, int y)
        {
            EntityDies?.Invoke(this, new(x, y));
        }

        private void InvokeFieldChanged(int x, int y, LowerFieldType lower)
        {
            FieldChanged?.Invoke(this, new(x, y, lower));
        }

        private void InvokeBombExplosion(int x, int y)
        {
            BombExplosion?.Invoke(this, new(x, y));
        }

        #endregion

        #region IDisposable

        private bool _disposed = false;

        ~GameModel() => Dispose(false);

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
                    _battleRoyaleTimer.Dispose();
                    _onePlayerDeadTimer.Dispose();
                    _bluePlayer.Dispose();
                    _redPlayer.Dispose();
                    foreach (Bomb bomb in _blueBombs)
                        bomb.Dispose();
                    foreach (Bomb bomb in _redBombs)
                        bomb.Dispose();
                    foreach (Monster monster in _monsters)
                        monster.Dispose();
                }

                _disposed = true;
            }
        }

        #endregion
    }
}
