﻿using Model.Utils;

using static Model.Utils.LowerFieldType;
using static Model.Utils.UpperFieldType;

namespace Model
{
    /// <summary>
    /// The GhostMonster is a subclass of the Monster class. It moves randomly but it can step on anything except the outer walls of the map.
    /// </summary>
    internal class GhostMonster : Monster
    {
        private LowerFieldType[] _onTopOfSomething = [InnerWall, Box, BlueBarrier, RedBarrier];

        /// <summary>
        /// Initializes a new instance of the GhostMonster class.
        /// </summary>
        /// <param name="x">Initial X position.</param>
        /// <param name="y">Initial Y position.</param>
        /// <param name="indexer">The indexer function of the map.</param>
        public GhostMonster(int x, int y, Func<int, int, (LowerFieldType, UpperFieldType)> indexer) : base(x, y, indexer)
        {
            _moveTimer.Interval = 2000;
            _moveTimer.Start();
            _canStepOnTheseLowers = [LowerFieldType.Empty, InnerWall, Box, BlueBarrier, RedBarrier, MoreBombsEffect, BiggerBombsEffect, DetonatorEffect, SkatesEffect, InvincibilityEffect, GhostEffect, BarrierEffect, Debuff];
            _canStepOnTheseUppers = [UpperFieldType.Empty, BluePlayer, RedPlayer];
        }

        /// \cond DOXYGEN_SHOULD_SKIP_THIS
        /// <summary>
        /// Only for testing! Initializes a new instance of the Model.GhostMonster class for testing.
        /// </summary>
        internal GhostMonster(int x, int y, Func<int, int, (LowerFieldType, UpperFieldType)> indexer, int interval) : this(x, y, indexer)
        {
            _moveTimer.Interval = interval;
        }
        /// \endcond

        protected override Direction? DoDecideDirectionToMove()
        {
            int fromX = X, fromY = Y;
            Direction direction = _direction;

            // if the monster is on top of something it can't change direction (only if it's necessary)
            bool canChangeDirection = !_onTopOfSomething.Contains(_mapIndexer(X, Y).lower);

            List<Direction> valids = GetValidDirections();

            if (valids.Count == 0)
            {
                // no valid moves
                return null;
            }

            if (valids.Count == 1)
            {
                // only one valid move
                direction = valids[0];
            }
            else
            {
                // several valid moves
                if (valids.Contains(direction))
                {
                    // current direction is valid
                    if (_random.Next(100) < 30 && canChangeDirection)
                    {
                        // chance of random movement (only if not on top of something)
                        valids.Remove(direction);
                        direction = valids[_random.Next(valids.Count)];
                    }
                }
                else
                {
                    // current direction is not valid, selecting random direction
                    direction = valids[_random.Next(valids.Count)];
                }
            }

            return direction;
        }

        #region IDisposable

        private bool _disposed = false;

        ~GhostMonster() => Dispose(false);

        protected override void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {

                }

                _disposed = true;

                base.Dispose(disposing);
            }
        }

        #endregion
    }
}
