﻿using Model.Utils;

using Timer = System.Timers.Timer;

namespace Model
{
    /// <summary>
    /// This is the abstract class for all the monsters in the game.
    /// </summary>
    internal abstract class Monster : IDisposable
    {
        #region Fields

        protected static Random _random = new();

        protected readonly Func<int, int, (LowerFieldType lower, UpperFieldType upper)> _mapIndexer;
        protected Direction _direction;
        protected Timer _moveTimer;

        #endregion

        #region Abstract Fields

        protected LowerFieldType[] _canStepOnTheseLowers = [];
        protected UpperFieldType[] _canStepOnTheseUppers = [];

        #endregion

        #region Properties

        /// <summary>
        /// The current X position of the monster.
        /// </summary>
        public int X { get; protected set; }

        /// <summary>
        /// The current Y position of the monster.
        /// </summary>
        public int Y { get; protected set; }

        #endregion

        #region Events

        /// <summary>
        /// Occures when the monster wants to move to a new position.
        /// </summary>
        public event EventHandler<MoveEventArgs>? MonsterWantsToMove;

        #endregion

        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="Monster"/> class.
        /// </summary>
        /// <param name="x">Initial x position.</param>
        /// <param name="y">Initial y position.</param>
        /// <param name="indexer">The indexer of the map.</param>
        protected Monster(int x, int y, Func<int, int, (LowerFieldType, UpperFieldType)> indexer)
        {
            _mapIndexer = indexer;
            X = x;
            Y = y;
            _direction = (Direction)_random.Next(4);

            _moveTimer = new();
            _moveTimer.Elapsed += (_, _) => PerformMove();
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// If the move was successful, this method should be called. It changes the position of the monster.
        /// </summary>
        /// <param name="x">New X position.</param>
        /// <param name="y">New Y position.</param>
        /// <param name="direction">The direction the monster is facing. Important for the textures.</param>
        public void SuccessfulMove(int x, int y, Direction direction)
        {
            X = x;
            Y = y;
            _direction = direction;
        }

        /// <summary>
        /// If the move was unsuccessful, this method should be called. It will try to move again.
        /// </summary>
        public void UnsuccessfulMove()
        {
            PerformMove();
        }

        /// <summary>
        /// If the monster died, this method should be called. It stops the timer.
        /// </summary>
        public void MonsterDied()
        {
            _moveTimer.Stop();
        }

        #endregion

        #region Protected Methods

        protected List<Direction> GetValidDirections()
        {
            List<Direction> valids = new List<Direction>();

            foreach (Direction dir in Enum.GetValues(typeof(Direction)))
            {
                switch (dir)
                {
                    case Direction.Up:
                        if (IsValid(X, Y - 1)) valids.Add(Direction.Up);
                        break;
                    case Direction.Down:
                        if (IsValid(X, Y + 1)) valids.Add(Direction.Down);
                        break;
                    case Direction.Left:
                        if (IsValid(X - 1, Y)) valids.Add(Direction.Left);
                        break;
                    case Direction.Right:
                        if (IsValid(X + 1, Y)) valids.Add(Direction.Right);
                        break;
                }
            }
            return valids;
        }

        protected bool IsValid(int x, int y)
        {
            return _canStepOnTheseLowers.Contains(_mapIndexer(x, y).lower) && _canStepOnTheseUppers.Contains(_mapIndexer(x, y).upper);
        }

        protected Direction? DirectionToClosestEnemy()
        {
            int startX = X;
            int startY = Y;


            Queue<(int, int)> queue = new Queue<(int, int)>();
            queue.Enqueue((startX, startY));

            HashSet<(int, int)> visited = new HashSet<(int, int)>();
            visited.Add((X, Y));

            Dictionary<(int, int), (int, int)> parent = new Dictionary<(int, int), (int, int)>();

            while (queue.Count > 0)
            {
                (int x, int y) = queue.Dequeue();

                if (_mapIndexer(x, y).upper == UpperFieldType.RedPlayer || _mapIndexer(x, y).upper == UpperFieldType.BluePlayer)
                {
                    return GetDirectionToTarget(startX, startY, x, y, parent);
                }

                foreach (Direction dir in Enum.GetValues(typeof(Direction)))
                {
                    int newX = x, newY = y;
                    switch (dir)
                    {
                        case Direction.Up:
                            newY--;
                            break;
                        case Direction.Down:
                            newY++;
                            break;
                        case Direction.Left:
                            newX--;
                            break;
                        case Direction.Right:
                            newX++;
                            break;
                    }
                    if (IsValid(newX, newY) && !visited.Contains((newX, newY)))
                    {
                        queue.Enqueue((newX, newY));
                        visited.Add((newX, newY));
                        parent[(newX, newY)] = (x, y);
                    }
                }
            }
            return null;
        }

        protected Direction GetDirectionToTarget(int startX, int startY, int targetX, int targetY, Dictionary<(int, int), (int, int)> parent)
        {
            (int, int) current = (targetX, targetY);
            while (parent[current].Item1 != startX || parent[current].Item2 != startY)
            {
                current = parent[current];
            }
            // how to get the direction from start to current
            int dx = current.Item1 - startX; // -1 left, 1 right, 0 same
            int dy = current.Item2 - startY; // -1 up, 1 down, 0 same
            if (dx == 0 && dy == 0)
            {
                throw new Exception("Error in GetDirectionToTarget, dx and dy both zero");
            }
            if (dx != 0 && dy != 0)
            {
                throw new Exception("Error in GetDirectionToTarget, dx and dy both non-zero");
            }
            if (dx == 1) return Direction.Right;
            if (dx == -1) return Direction.Left;
            if (dy == 1) return Direction.Down;
            return Direction.Up;
        }

        #endregion

        #region Template Methods

        protected void PerformMove()
        {
            int fromx = X, fromy = Y;
            Direction? dir = DoDecideDirectionToMove();

            if (dir == null)
                return;

            Direction toMove = (Direction)dir;

            int tox = X, toy = Y;
            switch (toMove)
            {
                case Direction.Up:
                    --toy; break;
                case Direction.Down:
                    ++toy; break;
                case Direction.Left:
                    --tox; break;
                case Direction.Right:
                    ++tox; break;
            }

            if (!IsValid(tox, toy))
            {
                PerformMove();
                return;
            }

            InvokeMonsterWantsToMove(fromx, fromy, tox, toy, toMove);
        }

        #endregion

        #region Primitive Methods

        protected abstract Direction? DoDecideDirectionToMove();

        #endregion

        #region Event Triggers

        protected void InvokeMonsterWantsToMove(int fromX, int fromY, int toX, int toY, Direction direction)
        {
            MonsterWantsToMove?.Invoke(this, new(fromX, fromY, toX, toY, direction));
        }

        #endregion

        #region Static Methods

        /// <summary>
        /// Returns a random monster and its enum type.
        /// </summary>
        /// <param name="x">Initial X position.</param>
        /// <param name="y">Initial Y position.</param>
        /// <param name="indexer">The indexer function of the map.</param>
        /// <returns>The freshly created monster and its enum type.</returns>
        public static (UpperFieldType, Monster) RandomMonsterFactory(int x, int y, Func<int, int, (LowerFieldType, UpperFieldType)> indexer)
        {
            int num = _random.Next(100);
            if (num < 25)
                return (UpperFieldType.BasicMonster, new BasicMonster(x, y, indexer));
            if (num < 50)
                return (UpperFieldType.GhostMonster, new GhostMonster(x, y, indexer));
            if (num < 75)
                return (UpperFieldType.SmartMonster, new SmartMonster(x, y, indexer));

            return (UpperFieldType.ClumsyMonster, new ClumsyMonster(x, y, indexer));
        }

        #endregion

        #region IDisposable

        private bool _disposed = false;

        ~Monster() => Dispose(false);

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
                    _moveTimer.Dispose();
                }

                _disposed = true;
            }
        }

        #endregion
    }
}
