﻿using Model.Utils;

using Timer = System.Timers.Timer;

namespace Model
{
    /// <summary>
    /// Represents a player in the game. It can calculate its next move, apply buffs and debuffs, and handle its own death.
    /// </summary>
    internal class Player : IDisposable
    {
        #region Constants

        private static readonly LowerFieldType[] canLowerBasic = [LowerFieldType.Empty, LowerFieldType.MoreBombsEffect, LowerFieldType.BiggerBombsEffect, LowerFieldType.DetonatorEffect, LowerFieldType.SkatesEffect, LowerFieldType.InvincibilityEffect, LowerFieldType.GhostEffect, LowerFieldType.BarrierEffect, LowerFieldType.Debuff];
        private static readonly UpperFieldType[] canUpperBasic = [UpperFieldType.Empty];
        private static readonly UpperFieldType[] deadlyBasic = [UpperFieldType.BasicMonster, UpperFieldType.ClumsyMonster, UpperFieldType.SmartMonster, UpperFieldType.GhostMonster, UpperFieldType.Explosion];
        private static readonly LowerFieldType[] canLowerGhost = [LowerFieldType.Empty, LowerFieldType.InnerWall, LowerFieldType.Box, LowerFieldType.BlueBarrier, LowerFieldType.RedBarrier, LowerFieldType.Bomb, LowerFieldType.MoreBombsEffect, LowerFieldType.BiggerBombsEffect, LowerFieldType.DetonatorEffect, LowerFieldType.SkatesEffect, LowerFieldType.InvincibilityEffect, LowerFieldType.GhostEffect, LowerFieldType.BarrierEffect, LowerFieldType.Debuff];
        private static readonly UpperFieldType[] canUpperGhost = canUpperBasic;
        private static readonly UpperFieldType[] deadlyGhost = deadlyBasic;
        private static readonly LowerFieldType[] canLowerInvincible = canLowerBasic;
        private static readonly UpperFieldType[] canUpperInvincible = canUpperBasic;
        private static readonly UpperFieldType[] deadlyInvincible = [];
        private static readonly LowerFieldType[] canLowerBoth = canLowerGhost;
        private static readonly UpperFieldType[] canUpperBoth = canUpperBasic;
        private static readonly UpperFieldType[] deadlyBoth = [];
        private static readonly LowerFieldType[] allEffects = [LowerFieldType.Debuff, LowerFieldType.MoreBombsEffect, LowerFieldType.BiggerBombsEffect, LowerFieldType.DetonatorEffect, LowerFieldType.SkatesEffect, LowerFieldType.BarrierEffect, LowerFieldType.InvincibilityEffect, LowerFieldType.GhostEffect];

        #endregion

        #region Fields

        private readonly Func<int, int, (LowerFieldType, UpperFieldType)> _mapIndexer;
        private int _slowBuffInterval;
        private int _invincibilityTime;
        private int _ghostTime;

        private Random _random = new Random();

        private Timer _moveTimer;

        private Timer _smallBombsTimer;
        private Timer _cantPlaceBombsTimer;
        private Timer _instantBombsTimer;
        private Timer _invincibilityTimer;
        private Timer _ghostTimer;

        #endregion

        #region Properties

        /// <summary>X position on the map.</summary>
        public int X { get; private set; }

        /// <summary>Y position on the map.</summary>
        public int Y { get; private set; }
        
        /// <summary>If the player can move in the current tick.</summary>
        public bool CanMove { get; private set; } = true;
        
        /// <summary>In each tick, a player can move once. This is the interval between two moves.</summary>
        public double MoveInterval { get => _moveTimer.Interval; }
        
        /// <summary>True if the Player has already died. Usually we return in functions if the player is dead.</summary>
        public bool IsDead { get; private set; } = false;
        
        /// <summary>The maximum number of bombs the player can place.</summary>
        public int MaxBombCount { get; private set; } = 1;
        
        /// <summary>The maximum range of the bombs the player can place.</summary>
        public int MaxBombRange { get; private set; } = 2;
        
        /// <summary>True if the player has a detonator. If the player has placed down all of their bombs, they can detonate them.</summary>
        public bool HasDetonator { get; private set; } = false;
        
        /// <summary>True if the player has skates. If the player has skates, they can move twice as fast.</summary>
        public bool HasSkates { get; private set; } = false;
        
        /// <summary>True if the player is invincible. If the player is invincible, they can't step on monsters, monsters can't step on them and the explosion stops before them.</summary>
        public bool IsInvincible { get; private set; } = false;
        
        /// <summary>True if the player is a ghost. If the player is a ghost, they can step on everything except for the outer walls.</summary>
        public bool IsGhost { get; private set; } = false;
        
        /// <summary>The maximum number of barriers a player can place.</summary>
        public int MaxBarrierCount { get; private set; } = 0;
        
        /// <summary>The number of barriers placed by the player. It can change if a barrier is destroyed.</summary>
        public int PlacedBarrierCount { get; private set; } = 0;
        
        /// <summary>True if the player can place a barrier.</summary>
        public bool CanPlaceBarrier { get => PlacedBarrierCount < MaxBarrierCount; }
        
        /// <summary>True if the player has just placed a barrier and hasn't moved since then.</summary>
        public bool JustPlacedBarrier { get; private set; } = false;
        
        /// <summary>True if the player has the small bombs debuff. The radius is 1.</summary>
        public bool HasSmallBombs { get; private set; } = false;
        
        /// <summary>True if the player has the cant place bombs debuff.</summary>
        public bool CantPlaceBombs { get; private set; } = false;
        
        /// <summary>True if the player has the instant bombs debuff. If the player has available bombs, they will be placed instantly.</summary>
        public bool InstantBombs { get; private set; } = false;

        #endregion

        #region Events

        /// <summary>
        /// Occurs when a player's special power changes.
        /// </summary>
        public event EventHandler<StatusEventArgs>? StatusChanged;
        /// <summary>
        /// Occurs when the invincibility buff is over.
        /// </summary>
        public event EventHandler? InvincibilityBuffIsOver;
        /// <summary>
        /// Occurs when the ghost buff is over.
        /// </summary>
        public event EventHandler? GhostBuffIsOver;
        /// <summary>
        /// Occurs when the player dies.
        /// </summary>
        public event EventHandler? PlayerDied;

        #endregion

        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="Player"/> class.
        /// </summary>
        /// <param name="x">Initial X position on the map.</param>
        /// <param name="y">Initial Y position on the map.</param>
        /// <param name="indexer">The indexer function of the map.</param>
        public Player(int x, int y, Func<int, int, (LowerFieldType, UpperFieldType)> indexer)
        {
            _mapIndexer = indexer;
            X = x;
            Y = y;
            _moveTimer = new Timer(500);
            _moveTimer.AutoReset = false;
            _moveTimer.Elapsed += (sender, e) => CanMove = true;
            _smallBombsTimer = new Timer(15000);
            _smallBombsTimer.AutoReset = false;
            _smallBombsTimer.Elapsed += SmallBombsDebuffOver;
            _cantPlaceBombsTimer = new Timer(15000);
            _cantPlaceBombsTimer.AutoReset = false;
            _cantPlaceBombsTimer.Elapsed += CantPlaceBombsDebuffOver;
            _instantBombsTimer = new Timer(15000);
            _instantBombsTimer.AutoReset = false;
            _instantBombsTimer.Elapsed += InstantBombsDebuffOver;
            _slowBuffInterval = 10000;

            _invincibilityTimer = new Timer(1000);
            _invincibilityTimer.Elapsed += (sender, e) => InvincibilityTick();

            _ghostTimer = new Timer(1000);
            _ghostTimer.Elapsed += (sender, e) => GhostTick();
        }

        /// \cond DOXYGEN_SHOULD_SKIP_THIS
        /// <summary>
        /// Only for testing! Initializes a new instance of the Model.Player class for testing.
        /// </summary>
        internal Player(int x, int y, Func<int, int, (LowerFieldType, UpperFieldType)> indexer, int interval) : this(x, y, indexer)
        {
            _moveTimer.Interval = interval;
            _smallBombsTimer.Interval = interval;
            _cantPlaceBombsTimer.Interval = interval;
            _instantBombsTimer.Interval = interval;
            _invincibilityTimer.Interval = interval;
            _ghostTimer.Interval = interval;
            _slowBuffInterval = interval;
        }
        /// \endcond

        #endregion

        #region Public Methods

        /// <summary>
        /// The player calculate its next move based on given direction.
        /// </summary>
        /// <param name="direction">The direction the player wants to move.</param>
        /// <returns>True if the player can move in that direction.</returns>
        public bool PlayerMoves(Direction direction)
        {
            if (!CanMove)
            {
                return false;
            }
            if (IsDead)
            {
                return false;
            }

            LowerFieldType[] canStepOnTheseLowers;
            UpperFieldType[] canStepOnTheseUppers;
            UpperFieldType[] deadlyIfStepsOnIt;
            if (IsGhost)
            {
                if (IsInvincible)
                {
                    canStepOnTheseLowers = canLowerBoth;
                    canStepOnTheseUppers = canUpperBoth;
                    deadlyIfStepsOnIt = deadlyBoth;
                }
                else
                {
                    canStepOnTheseLowers = canLowerGhost;
                    canStepOnTheseUppers = canUpperGhost;
                    deadlyIfStepsOnIt = deadlyGhost;
                }
            }
            else
            {
                if (IsInvincible)
                {
                    canStepOnTheseLowers = canLowerInvincible;
                    canStepOnTheseUppers = canUpperInvincible;
                    deadlyIfStepsOnIt = deadlyInvincible;
                }
                else
                {
                    canStepOnTheseLowers = canLowerBasic;
                    canStepOnTheseUppers = canUpperBasic;
                    deadlyIfStepsOnIt = deadlyBasic;
                }
            }

            // success means the player moved -> a) to a new position, b) died
            // !success means the player didn't move -> can't move, dead, can't move to that direction
            bool success = false;
            int fromX = X;
            int fromY = Y;
            LowerFieldType l;
            UpperFieldType u;
            switch (direction)
            {
                case Direction.Up:
                    (l, u) = _mapIndexer(X, Y - 1);
                    if (canStepOnTheseUppers.Contains(u) && canStepOnTheseLowers.Contains(l))
                    {
                        CheckEffects(l);
                        Y--;
                        success = true;
                    }
                    else if (deadlyIfStepsOnIt.Contains(u))
                    {
                        PlayerDies();
                        success = true;
                    }
                    break;
                case Direction.Down:
                    (l, u) = _mapIndexer(X, Y + 1);
                    if (canStepOnTheseUppers.Contains(u) && canStepOnTheseLowers.Contains(l))
                    {
                        CheckEffects(l);
                        Y++;
                        success = true;
                    }
                    else if (deadlyIfStepsOnIt.Contains(u))
                    {
                        PlayerDies();
                        success = true;
                    }
                    break;
                case Direction.Left:
                    (l, u) = _mapIndexer(X - 1, Y);
                    if (canStepOnTheseUppers.Contains(u) && canStepOnTheseLowers.Contains(l))
                    {
                        CheckEffects(l);
                        X--;
                        success = true;
                    }
                    else if (deadlyIfStepsOnIt.Contains(u))
                    {
                        PlayerDies();
                        success = true;
                    }
                    break;
                case Direction.Right:
                    (l, u) = _mapIndexer(X + 1, Y);
                    if (canStepOnTheseUppers.Contains(u) && canStepOnTheseLowers.Contains(l))
                    {
                        CheckEffects(l);
                        X++;
                        success = true;
                    }
                    else if (deadlyIfStepsOnIt.Contains(u))
                    {
                        PlayerDies();
                        success = true;
                    }
                    break;
            }
            if (success)
            {
                CanMove = false;
                JustPlacedBarrier = false;
                if (!IsDead)
                {
                    _moveTimer.Start();
                }
            }
            return success;
        }

        /// <summary>
        /// Stopping the timers and other values if the player has died. Invokes the PlayerDied event.
        /// </summary>
        public void PlayerDies()
        {
            IsDead = true;
            _moveTimer.Stop();
            _cantPlaceBombsTimer.Stop();
            _instantBombsTimer.Stop();
            _invincibilityTimer.Stop();
            _ghostTimer.Stop();
            StatusChanged?.Invoke(this, new(SpecialPower.Invincibility, false));
            StatusChanged?.Invoke(this, new(SpecialPower.Ghost, false));
            CanMove = false;
            PlayerDied?.Invoke(this, EventArgs.Empty);
        }

        /// <summary>
        /// Applies the More Bombs Buff to the player.
        /// </summary>
        public void ApplyMoreBombsBuff()
        {
            MaxBombCount++;
        }

        /// <summary>
        /// Applies the Bigger Bombs Buff to the player.
        /// </summary>
        public void ApplyBiggerBombsBuff()
        {
            MaxBombRange++;
        }

        /// <summary>
        /// Applies the Detonator Buff to the player.
        /// </summary>
        public void ApplyDetonatorBuff()
        {
            HasDetonator = true;
        }

        /// <summary>
        /// Applies the Skates Buff to the player.
        /// </summary>
        public void ApplySkatesBuff()
        {
            if (!HasSkates)
            {
                HasSkates = true;
                _moveTimer.Interval /= 2;
            }
        }

        /// <summary>
        /// Applies the Invincibility Buff to the player.
        /// </summary>
        public void ApplyInvincibilityBuff()
        {
            IsInvincible = true;
            _invincibilityTimer.Start();
            _invincibilityTime = 10;
            StatusChanged?.Invoke(this, new StatusEventArgs(SpecialPower.Invincibility, true));
        }

        /// <summary>
        /// Applies the Ghost Buff to the player.
        /// </summary>
        public void ApplyGhostBuff()
        {
            IsGhost = true;
            _ghostTimer.Start();
            _ghostTime = 10;
            StatusChanged?.Invoke(this, new StatusEventArgs(SpecialPower.Ghost, true));
        }

        /// <summary>
        /// Applies the Barrier Buff to the player.
        /// </summary>
        public void ApplyBarrierBuff()
        {
            MaxBarrierCount += 3;
        }

        /// <summary>
        /// The player placed a barrier. Increases the PlacedBarrierCount and sets JustPlacedBarrier to true.
        /// </summary>
        public void BarrierPlaced()
        {
            PlacedBarrierCount++;
            JustPlacedBarrier = true;
        }

        /// <summary>
        /// A barrier was destroyed. Decreases the PlacedBarrierCount.
        /// </summary>
        public void BarrierDestroyed()
        {
            PlacedBarrierCount--;
        }

        /// <summary>
        /// Applies the Slow Debuff to the player.
        /// </summary>
        public void ApplySlowDebuff()
        {
            if (_moveTimer.Interval < 1500)
            {
                StackableEffect debuff = new StackableEffect(_slowBuffInterval, () => _moveTimer.Interval *= 1.2,
                    () => _moveTimer.Interval /= 1.2);
            }
        }

        /// <summary>
        /// Applies the Small Bombs Debuff to the player.
        /// </summary>
        public void ApplySmallBombsDebuff()
        {
            HasSmallBombs = true;
            _smallBombsTimer.Stop();
            _smallBombsTimer.Start();
        }

        /// <summary>
        /// Applies the Cant Place Bombs Debuff to the player.
        /// </summary>
        public void ApplyCantPlaceBombsDebuff()
        {
            CantPlaceBombs = true;
            _cantPlaceBombsTimer.Stop();
            _cantPlaceBombsTimer.Start();
        }

        /// <summary>
        /// Applies the Instant Bombs Debuff to the player.
        /// </summary>
        public void ApplyInstantBombsDebuff()
        {
            InstantBombs = true;
            _instantBombsTimer.Stop();
            _instantBombsTimer.Start();
        }

        #endregion

        #region Private methods

        private void CheckEffects(LowerFieldType l)
        {
            if (allEffects.Contains(l))
            {
                switch (l)
                {
                    case LowerFieldType.MoreBombsEffect:
                        ApplyMoreBombsBuff();
                        break;
                    case LowerFieldType.BiggerBombsEffect:
                        ApplyBiggerBombsBuff();
                        break;
                    case LowerFieldType.DetonatorEffect:
                        ApplyDetonatorBuff();
                        break;
                    case LowerFieldType.SkatesEffect:
                        ApplySkatesBuff();
                        break;
                    case LowerFieldType.BarrierEffect:
                        ApplyBarrierBuff();
                        break;
                    case LowerFieldType.Debuff:
                        int num = _random.Next(100);
                        if (num < 25)
                            ApplySlowDebuff();
                        else if (num < 50)
                            ApplySmallBombsDebuff();
                        else if (num < 75)
                            ApplyCantPlaceBombsDebuff();
                        else
                            ApplyInstantBombsDebuff();
                        break;
                    case LowerFieldType.InvincibilityEffect:
                        ApplyInvincibilityBuff();
                        break;
                    case LowerFieldType.GhostEffect:
                        ApplyGhostBuff();
                        break;
                    default:
                        throw new ArgumentException("Unimplemented effect!");
                }
            }
        }

        private void SmallBombsDebuffOver(object? sender, System.Timers.ElapsedEventArgs e)
        {
            HasSmallBombs = false;
        }

        private void CantPlaceBombsDebuffOver(object? sender, System.Timers.ElapsedEventArgs e)
        {
            CantPlaceBombs = false;
        }

        private void InstantBombsDebuffOver(object? sender, System.Timers.ElapsedEventArgs e)
        {
            InstantBombs = false;
        }

        private void InvincibilityTick()
        {
            if (--_invincibilityTime > 6)
                return;

            if (_invincibilityTime % 2 == 0)
            {
                StatusChanged?.Invoke(this, new StatusEventArgs(SpecialPower.Invincibility, false));
            }
            else
            {
                StatusChanged?.Invoke(this, new StatusEventArgs(SpecialPower.Invincibility, true));
            }
            if (_invincibilityTime == 0)
            {
                _invincibilityTimer.Stop();
                IsInvincible = false;
                InvincibilityBuffIsOver?.Invoke(this, EventArgs.Empty);
            }
        }

        private void GhostTick()
        {
            if (--_ghostTime > 6)
                return;

            if (_ghostTime % 2 == 0)
            {
                StatusChanged?.Invoke(this, new StatusEventArgs(SpecialPower.Ghost, false));
            }
            else
            {
                StatusChanged?.Invoke(this, new StatusEventArgs(SpecialPower.Ghost, true));
            }
            if (_ghostTime == 0)
            {
                _ghostTimer.Stop();
                IsGhost = false;
                GhostBuffIsOver?.Invoke(this, EventArgs.Empty);
            }
        }

        #endregion

        #region IDisposable

        private bool _disposed = false;

        ~Player() => Dispose(false);

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
                    _smallBombsTimer.Dispose();
                    _cantPlaceBombsTimer.Dispose();
                    _instantBombsTimer.Dispose();
                    _invincibilityTimer.Dispose();
                    _ghostTimer.Dispose();
                }

                _disposed = true;
            }
        }

        #endregion
    }
}
