﻿using Model.Utils;
using System.Globalization;
using System.Windows;
using System.Windows.Data;
using System.Windows.Media.Imaging;

namespace View
{
    /// <summary>
    /// Converts LowerFieldTypes or Tuple&lt;UpperFieldType, Direction&gt;-s to their corresponding images. 
    /// </summary>
    internal class ImageConverter : IValueConverter
    {
        private Random random = new Random();

        /// <summary>
        /// Converts a LowerFieldTypes or Tuple&lt;UpperFieldType, Direction&gt; to their corresponding images.
        /// </summary>
        /// <param name="value">The LowerFieldType or Tuple&lt;UpperFieldType, Direction&gt;.</param>
        /// <param name="targetType">Ignored. It is expected to be a BitmapImage or one of it's bases.</param>
        /// <param name="parameter">Ignored.</param>
        /// <param name="culture">Ignored.</param>
        /// <returns>The BitmapImage that corresponds to the input value.</returns>
        /// <exception cref="ArgumentException">Thrown when the <paramref name="value"/> is not one of LowerFieldType or Tuple&lt;UpperFieldType, Direction&gt;, or if an unknown enum value is received.</exception>
        public object Convert(object value, Type targetType, object parameter, CultureInfo culture)
        {
            string path = value switch
            {
                LowerFieldType lower => lower switch
                {
                    LowerFieldType.Empty               => string.Empty,
                    LowerFieldType.Bomb                => "cherrybomb.png",
                    LowerFieldType.InnerWall           => "wallnut_basic.png",
                    LowerFieldType.OuterWall           => "water_tile.png",
                    LowerFieldType.Box                 => "vase.png",
                    LowerFieldType.BlueBarrier         => "barrier_blue.png",
                    LowerFieldType.RedBarrier          => "barrier_red.png",
                    LowerFieldType.MoreBombsEffect     => "cherrybomb_plus.png",
                    LowerFieldType.BiggerBombsEffect   => "cherrybomb_big.png",
                    LowerFieldType.DetonatorEffect     => "detonator.png",
                    LowerFieldType.SkatesEffect        => "coffee_bean.png",
                    LowerFieldType.InvincibilityEffect => "shield.png",
                    LowerFieldType.GhostEffect         => "lufi.png",
                    LowerFieldType.BarrierEffect       => "barrier_effect.png",
                    LowerFieldType.Debuff              => "hypno_shroom.png",
                    _ => throw new ArgumentException($"Unknown LowerFieldType value received: {lower}"),
                },

                Tuple<UpperFieldType, Direction> upper => upper.Item1 switch
                {
                    UpperFieldType.Empty => string.Empty,
                    UpperFieldType.BluePlayer => upper.Item2 switch
                    {
                        Direction.Up =>    "blue_player_up.png",
                        Direction.Down =>  "blue_player_down.png",
                        Direction.Left =>  "blue_player_left.png",
                        Direction.Right => "blue_player_right.png",
                        _ => throw new ArgumentException($"Unknown Direction value received: {upper.Item2}"),
                    },
                    UpperFieldType.RedPlayer => upper.Item2 switch
                    {
                        Direction.Up =>    "red_player_up.png",
                        Direction.Down =>  "red_player_down.png",
                        Direction.Left =>  "red_player_left.png",
                        Direction.Right => "red_player_right.png",
                        _ => throw new ArgumentException($"Unknown Direction value received: {upper.Item2}"),
                    },
                    UpperFieldType.BasicMonster => upper.Item2 switch
                    {
                        Direction.Up =>    "basic_monster_up.png",
                        Direction.Down =>  "basic_monster_down.png",
                        Direction.Left =>  "basic_monster_left.png",
                        Direction.Right => "basic_monster_right.png",
                        _ => throw new ArgumentException($"Unknown Direction value received: {upper.Item2}"),
                    },
                    UpperFieldType.GhostMonster => upper.Item2 switch
                    {
                        Direction.Up =>    "ghost_monster_up.png",
                        Direction.Down =>  "ghost_monster_down.png",
                        Direction.Left =>  "ghost_monster_left.png",
                        Direction.Right => "ghost_monster_right.png",
                        _ => throw new ArgumentException($"Unknown Direction value received: {upper.Item2}"),
                    },
                    UpperFieldType.SmartMonster => upper.Item2 switch
                    {
                        Direction.Up =>    "smart_monster_up.png",
                        Direction.Down =>  "smart_monster_down.png",
                        Direction.Left =>  "smart_monster_left.png",
                        Direction.Right => "smart_monster_right.png",
                        _ => throw new ArgumentException($"Unknown Direction value received: {upper.Item2}"),
                    },
                    UpperFieldType.ClumsyMonster => upper.Item2 switch
                    {
                        Direction.Up =>    "clumsy_monster_up.png",
                        Direction.Down =>  "clumsy_monster_down.png",
                        Direction.Left =>  "clumsy_monster_left.png",
                        Direction.Right => "clumsy_monster_right.png",
                        _ => throw new ArgumentException($"Unknown Direction value received: {upper.Item2}"),
                    },
                    UpperFieldType.Explosion => GetRandomBombPath(),
                    _ => throw new ArgumentException($"Unknown UpperFieldType value received: {upper.Item1}")
                },

                _ => throw new ArgumentException("Argument was not LowerFieldType or Tuple<UpperFieldType, Direction>!", nameof(value)),
            };

            if (path == string.Empty)
                return DependencyProperty.UnsetValue;

            path = $"Media\\{path}";
            return new BitmapImage(new Uri(path, UriKind.Relative));
        }

        /// <summary>
        /// Not implemented. Throws NotImplementedException.
        /// </summary>
        /// <exception cref="NotImplementedException">Thrown every time.</exception>
        public object ConvertBack(object value, Type targetType, object parameter, CultureInfo culture)
        {
            throw new NotImplementedException();
        }

        private string GetRandomBombPath()
        {
            return random.Next(12) switch
            {
                0 => "bomb_gt.png",
                1 => "bomb_bamm.png",
                2 => "bomb_bumm.png",
                3 => "bomb_puff.png",
                4 => "bomb_kapow.png",
                5 => "bomb_pow.png",
                _ => "powie.png",
            };
        }
    }
}
