﻿using System.Windows.Input;

namespace ViewModel
{
    /// <summary>
    /// Encapsulates a command, which can also have execution disabled.
    /// </summary>
    public class DelegateCommand : ICommand
    {
        private readonly Action<object?> _execute;
        private readonly Func<object?, bool>? _canExecute;

        /// <summary>
        /// Occurs when the executability of this command changed.
        /// </summary>
        public event EventHandler? CanExecuteChanged;

        /// <summary>
        /// Initializes a new instance of the ViewModel.DelegateCommand class.
        /// </summary>
        /// <param name="execute">The action to be executed.</param>
        public DelegateCommand(Action<object?> execute) : this(execute, null) { }

        /// <summary>
        /// Initializes a new instance of the ViewModel.DelegateCommand class.
        /// </summary>
        /// <param name="execute">The action to be executed.</param>
        /// <param name="canExecute">The predicate that determines if the action can be executed.</param>
        public DelegateCommand(Action<object?> execute, Func<object?, bool>? canExecute)
        {
            _execute = execute;
            _canExecute = canExecute;
        }

        /// <summary>
        /// Determines if the command can be executed.
        /// </summary>
        /// <param name="parameter">The parameter of the action.</param>
        /// <returns>Returns true if the command can be executed, false otherwise.</returns>
        public bool CanExecute(object? parameter)
        {
            return _canExecute == null ? true : _canExecute(parameter);
        }

        /// <summary>
        /// Executes the command.
        /// </summary>
        /// <param name="parameter">The parameter of the action.</param>
        public void Execute(object? parameter)
        {
            if (!CanExecute(parameter))
                throw new InvalidOperationException("Command execution is disabled!");

            _execute(parameter);
        }

        /// <summary>
        /// Raises the event that notifies subscribed classes that the executability changed.
        /// </summary>
        public void RaiseCanExecuteChanged()
        {
            CanExecuteChanged?.Invoke(this, EventArgs.Empty);
        }
    }
}
