using Model;
using Model.Utils;
using System.Collections.ObjectModel;

namespace ViewModel
{
    /// <summary>
    /// Provides bindable data and handles the model for the Bomberman game.
    /// </summary>
    public class MainViewModel : ViewModelBase
    {
        #region Fields

        private readonly GameModel _model;
        private ObservableCollection<GameField> _fields;

        private int _layoutNumber;
        private int _pointsToWin;
        private int _battleRoyalTimeLeft;
        private bool _blueIsGhost;
        private bool _redIsGhost;
        private bool _blueIsInvincible;
        private bool _redIsInvincible;

        /// <summary>
        /// Stores which page are we on. Should only be manipulated with SwapToMenuPage or SwapToGamePage!
        /// </summary>
        private bool _isMenu;
        private bool _isFaded;

        #endregion

        #region Properties

        /// <summary>Gets or sets the value that indicates if blue is moving up. Should be set true and false respectively.</summary>
        public bool BlueUpPressed { get; set; }

        /// <summary>Gets or sets the value that indicates if blue is moving down. Should be set true and false respectively.</summary>
        public bool BlueDownPressed { get; set; }

        /// <summary>Gets or sets the value that indicates if blue is moving left. Should be set true and false respectively.</summary>
        public bool BlueLeftPressed { get; set; }

        /// <summary>Gets or sets the value that indicates if blue is moving right. Should be set true and false respectively.</summary>
        public bool BlueRightPressed { get; set; }

        /// <summary>Gets or sets the value that indicates if blue is placing a bomb. Should be set true and false respectively.</summary>
        public bool BlueBombPressed { get; set; }

        /// <summary>Gets or sets the value that indicates if blue has placed a bomb on this button press. MainViewModel sets it to true, should be set to false if the button is released.</summary>
        public bool BlueBombRepeat { get; set; }

        /// <summary>Gets or sets the value that indicates if blue is placing a barrier. Should be set true and false respectively.</summary>
        public bool BlueBarrierPressed { get; set; }

        /// <summary>Gets or sets the value that indicates if red is moving up. Should be set true and false respectively.</summary>
        public bool RedUpPressed { get; set; }

        /// <summary>Gets or sets the value that indicates if red is moving down. Should be set true and false respectively.</summary>
        public bool RedDownPressed { get; set; }

        /// <summary>Gets or sets the value that indicates if red is moving left. Should be set true and false respectively.</summary>
        public bool RedLeftPressed { get; set; }

        /// <summary>Gets or sets the value that indicates if red is moving right. Should be set true and false respectively.</summary>
        public bool RedRightPressed { get; set; }

        /// <summary>Gets or sets the value that indicates if red is placing a bomb. Should be set true and false respectively.</summary>
        public bool RedBombPressed { get; set; }

        /// <summary>Gets or sets the value that indicates if red has placed a bomb on this button press. MainViewModel sets it to true, should be set to false if the button is released.</summary>
        public bool RedBombRepeat { get; set; }

        /// <summary>Gets or sets the value that indicates if red is placing a barrier. Should be set true and false respectively.</summary>
        public bool RedBarrierPressed { get; set; }

        /// <summary>
        /// Gets the height of the game field.
        /// </summary>
        public int Height
        {
            get => _model.HEIGHT;
        }

        /// <summary>
        /// Gets the width of the game field.
        /// </summary>
        public int Width
        {
            get => _model.WIDTH;
        }

        /// <summary>
        /// Gets the ObservableCollection containing info on all tiles. Used for binding.
        /// </summary>
        public ObservableCollection<GameField> Fields
        {
            get => _fields;
            private set
            {
                if (_fields != value)
                {
                    _fields = value;
                    OnPropertyChanged();
                }
            }
        }

        /// <summary>
        /// Gets the number of points needed to win.
        /// </summary>
        public int PointsToWin
        {
            get => _pointsToWin;
            private set
            {
                if (_pointsToWin != value)
                {
                    _pointsToWin = value;
                    OnPropertyChanged();
                }
            }
        }

        /// <summary>
        /// Gets the number of points the blue player has.
        /// </summary>
        public int BluePoints
        {
            get => _model.BluePoints;
        }

        /// <summary>
        /// Gets the number of points the red player has.
        /// </summary>
        public int RedPoints
        {
            get => _model.RedPoints;
        }

        /// <summary>
        /// Gets the time remaining until the next battle royale step, is seconds.
        /// </summary>
        public int BattleRoyalTimeLeft
        {
            get => _battleRoyalTimeLeft;
            private set
            {
                if (_battleRoyalTimeLeft != value)
                {
                    _battleRoyalTimeLeft = value;
                    OnPropertyChanged();
                }
            }
        }

        /// <summary>
        /// Gets the value that indicates if the blue player is invincible.
        /// </summary>
        public bool BlueIsInvincible
        {
            get => _blueIsInvincible;
            private set
            {
                if (_blueIsInvincible != value)
                {
                    _blueIsInvincible = value;
                    OnPropertyChanged();
                }
            }
        }

        /// <summary>
        /// Gets the value that indicates if the red player is invicible.
        /// </summary>
        public bool RedIsInvincible
        {
            get => _redIsInvincible;
            private set
            {
                if (_redIsInvincible != value)
                {
                    _redIsInvincible = value;
                    OnPropertyChanged();
                }
            }
        }

        /// <summary>
        /// Gets the value that indicates if the blue player is a ghost.
        /// </summary>
        public bool BlueIsGhost
        {
            get => _blueIsGhost;
            private set
            {
                if (_blueIsGhost != value)
                {
                    _blueIsGhost = value;
                    OnPropertyChanged();
                }
            }
        }

        /// <summary>
        /// Gets the value that indicates if the red player is a ghost.
        /// </summary>
        public bool RedIsGhost
        {
            get => _redIsGhost;
            private set
            {
                if (_redIsGhost != value)
                {
                    _redIsGhost = value;
                    OnPropertyChanged();
                }
            }
        }

        /// <summary>
        /// Gets the value that indicates if we are on the menu page.
        /// </summary>
        public bool IsMenu
        {
            get => _isMenu;
        }

        /// <summary>
        /// Gets the value that indicates if we are on the game page.
        /// </summary>
        public bool IsGame
        {
            get => !_isMenu;
        }

        /// <summary>
        /// Gets the value that indicates if the screen is fading.
        /// </summary>
        public bool IsFaded
        {
            get => _isFaded;
            private set
            {
                if (_isFaded != value)
                {
                    _isFaded = value;
                    OnPropertyChanged();
                }
            }
        }

        /// <summary>
        /// Gets the value that indicates the current layout.
        /// </summary>
        public int LayoutNumber
        {
            get => _layoutNumber;
            private set
            {
                if (value != _layoutNumber && value is 1 or 2 or 3)
                {
                    _layoutNumber = value;
                    OnPropertyChanged();
                    OnPropertyChanged(nameof(LayoutOneSelected));
                    OnPropertyChanged(nameof(LayoutTwoSelected));
                    OnPropertyChanged(nameof(LayoutThreeSelected));
                }
            }
        }

        /// <summary>
        /// Gets the value that indicates if layout one is selected.
        /// </summary>
        public bool LayoutOneSelected
        {
            get => LayoutNumber == 1;
        }

        /// <summary>
        /// Gets the value that indicates if layout two is selected.
        /// </summary>
        public bool LayoutTwoSelected
        {
            get => LayoutNumber == 2;
        }

        /// <summary>
        /// Gets the value that indicates if layout three is selected.
        /// </summary>
        public bool LayoutThreeSelected
        {
            get => LayoutNumber == 3;
        }

        #endregion

        #region Commands

        /// <summary>Gets the command that works on the menu page and starts a new competition.</summary>
        public DelegateCommand NewCompetitionCommand { get; private set; }

        /// <summary>Gets the command that works on the menu page and exits the application.</summary>
        public DelegateCommand QuitApplicationCommand { get; private set; }

        /// <summary>Gets the command that works on the menu page and reduces the number of points to win.</summary>
        public DelegateCommand LessNumberOfPointsToWinCommand { get; private set; }

        /// <summary>Gets the command that works on the menu page and increments the number of points to win.</summary>
        public DelegateCommand MoreNumberOfPointsToWinCommand { get; private set; }

        /// <summary>Gets the command that works on the menu page and selects another layout.</summary>
        public DelegateCommand SelectedTableChangedCommand { get; private set; }

        /// <summary>Gets the command that works on the game page and ends the current competition.</summary>
        public DelegateCommand TerminateCompetitionCommand { get; private set; }

        #endregion

        #region Events

        /// <summary>
        /// Occurs when some code should be ran on the main thread (usually when view needs to be updated).
        /// </summary>
        public event EventHandler<ActionEventArgs>? ExecuteOnMainThread;
        /// <summary>
        /// Occurs when the application should close.
        /// </summary>
        public event EventHandler? QuitApplication;
        /// <summary>
        /// Occurs when a player has won the competition.
        /// </summary>
        public event EventHandler<PlayerEventArgs>? CompetitionOver;

        #endregion

        #region Constructors

        /// <summary>
        /// Initializes a new instance of the ViewModel.MainViewModel class.
        /// </summary>
        /// <param name="model">The model instance the view model should use.</param>
        public MainViewModel(GameModel model)
        {
            _isMenu = true;
            _layoutNumber = 1;
            _pointsToWin = 2;

            _model = model;
            _model.NewRound                += GameModel_NewRoundStarted;
            _model.EntityMoved             += GameModel_EntityMoved;
            _model.EntityDies              += GameModel_EntityDies;
            _model.FieldChanged            += GameModel_FieldChanged;
            _model.BombExplosion           += GameModel_BombExplosion;
            _model.CompetitionOver         += GameModel_CompetitionOver;
            _model.BattleRoyaleTimeChanged += GameModel_BattleRoyalTimeChanged;
            _model.PlayerStatusChanged     += GameModel_PlayerStatusChanged;

            _fields = new ObservableCollection<GameField>();
            for (int i = 0; i < Width; i++)
            {
                for (int j = 0; j < Height; j++)
                {
                    Fields.Add(new());
                }
            }

            Func<object?, bool> isMenuPredicate = _ => IsMenu;
            Func<object?, bool> isGamePredicate = _ => IsGame;

            NewCompetitionCommand          = new DelegateCommand(OnNewCompetition,       isMenuPredicate);
            LessNumberOfPointsToWinCommand = new DelegateCommand(OnLessPoints,           isMenuPredicate);
            MoreNumberOfPointsToWinCommand = new DelegateCommand(OnMorePoints,           isMenuPredicate);
            SelectedTableChangedCommand    = new DelegateCommand(OnSelectedTableChanged, isMenuPredicate);
            QuitApplicationCommand         = new DelegateCommand(OnQuitApplication,      isMenuPredicate);

            TerminateCompetitionCommand = new DelegateCommand(OnTerminateCompetition, isGamePredicate);
        }

        #endregion

        #region Private Methods

        private void OnLessPoints(object? param)
        {
            if (PointsToWin > 1)
            {
                PointsToWin--;
            }
        }

        private void OnMorePoints(object? param)
        {
            if (PointsToWin < 10)
            {
                PointsToWin++;
            }
        }

        private async void OnNewCompetition(object? param)
        {
            _ = SwapToGamePage();
            await Task.Delay(1000);
            _model.NewCompetitionStarts(_pointsToWin, LayoutNumber);
            OnPropertyChanged(nameof(PointsToWin));
        }

        private async void OnQuitApplication(object? param)
        {
            IsFaded = true;
            await Task.Delay(1000);
            QuitApplication?.Invoke(this, EventArgs.Empty);
        }

        private void OnSelectedTableChanged(object? param)
        {
            if (param is string p)
            {
                LayoutNumber = p switch
                {
                    "layout1" => 1,
                    "layout2" => 2,
                    "layout3" => 3,
                    _ => throw new ArgumentException("Received unknown layout!"),
                };
            }
            else
            {
                throw new ArgumentException("Argument was not string!", nameof(param));
            }
        }

        private async void OnTerminateCompetition(object? param)
        {
            _model.TerminateCompetition();
            await SwapToMenuPage();
        }

        private async Task HandleButtonInput()
        {
            await Task.Run(() =>
            {
                while (IsGame)
                {
                    int blueVertical    = BlueUpPressed   ? (BlueDownPressed  ? 0 : -1) : (BlueDownPressed  ? 1 : 0);
                    int blueHorizontal  = BlueLeftPressed ? (BlueRightPressed ? 0 : -1) : (BlueRightPressed ? 1 : 0);

                    if (blueVertical != 0)
                    {
                        if (blueVertical == 1)
                            _model.Move(PlayerType.Blue, Direction.Down);
                        else
                            _model.Move(PlayerType.Blue, Direction.Up);
                    }
                    else
                    {
                        if (blueHorizontal != 0)
                        {
                            if (blueHorizontal == 1)
                                _model.Move(PlayerType.Blue, Direction.Right);
                            else
                                _model.Move(PlayerType.Blue, Direction.Left);
                        }
                    }
                    if (BlueBombPressed && !BlueBombRepeat)
                    {
                        BlueBombRepeat = true;
                        _model.PlaceBomb(PlayerType.Blue);
                    }
                    if (BlueBarrierPressed)
                        _model.PlaceBarrier(PlayerType.Blue);

                    int redVertical = RedUpPressed ? (RedDownPressed ? 0 : -1) : (RedDownPressed ? 1 : 0);
                    int redHorizontal = RedLeftPressed ? (RedRightPressed ? 0 : -1) : (RedRightPressed ? 1 : 0);

                    if (redVertical != 0)
                    {
                        if (redVertical == 1)
                            _model.Move(PlayerType.Red, Direction.Down);
                        else
                            _model.Move(PlayerType.Red, Direction.Up);
                    }
                    else
                    {
                        if (redHorizontal != 0)
                        {
                            if (redHorizontal == 1)
                                _model.Move(PlayerType.Red, Direction.Right);
                            else
                                _model.Move(PlayerType.Red, Direction.Left);
                        }
                    }
                    if (RedBombPressed && !RedBombRepeat)
                    {
                        RedBombRepeat = true;
                        _model.PlaceBomb(PlayerType.Red);
                    }
                    if (RedBarrierPressed)
                        _model.PlaceBarrier(PlayerType.Red);
                }
            });
        }

        private async Task SwapToMenuPage()
        {
            if (IsMenu)
                return;

            _isMenu = true;

            InvokeExecuteOnMainThread(() =>
            {
                TerminateCompetitionCommand.RaiseCanExecuteChanged();

                IsFaded = true;
            });

            await Task.Delay(1000);

            InvokeExecuteOnMainThread(() =>
            {
                OnPropertyChanged(nameof(IsGame));
                OnPropertyChanged(nameof(IsMenu));

                NewCompetitionCommand.RaiseCanExecuteChanged();
                QuitApplicationCommand.RaiseCanExecuteChanged();
                LessNumberOfPointsToWinCommand.RaiseCanExecuteChanged();
                MoreNumberOfPointsToWinCommand.RaiseCanExecuteChanged();
                SelectedTableChangedCommand.RaiseCanExecuteChanged();

                IsFaded = false;
            });
        }

        private async Task SwapToGamePage()
        {
            if (IsGame)
                return;

            _isMenu = false;

            InvokeExecuteOnMainThread(() =>
            {
                NewCompetitionCommand.RaiseCanExecuteChanged();
                QuitApplicationCommand.RaiseCanExecuteChanged();
                LessNumberOfPointsToWinCommand.RaiseCanExecuteChanged();
                MoreNumberOfPointsToWinCommand.RaiseCanExecuteChanged();
                SelectedTableChangedCommand.RaiseCanExecuteChanged();

                IsFaded = true;
            });

            await Task.Delay(1000);

            InvokeExecuteOnMainThread(() =>
            {
                OnPropertyChanged(nameof(IsMenu));
                OnPropertyChanged(nameof(IsGame));

                IsFaded = false;

                TerminateCompetitionCommand.RaiseCanExecuteChanged();

                _ = HandleButtonInput();
            });
        }

        #endregion

        #region Private Event Methods

        private void GameModel_NewRoundStarted(object? sender, NewRoundEventArgs e)
        {
            InvokeExecuteOnMainThread(() =>
            {
                BlueIsInvincible = false;
                BlueIsGhost = false;
                RedIsInvincible = false;
                RedIsGhost = false;
                for (int x = 0; x < Width; x++)
                {
                    for (int y = 0; y < Height; y++)
                    {
                        Fields[y * Width + x].LowerLayer = e.LowerLayer[y, x];
                        Fields[y * Width + x].UpperLayer = e.UpperLayer[y, x];
                        Fields[y * Width + x].Dir = Direction.Down;
                    }
                }
                OnPropertyChanged(nameof(Fields));
                OnPropertyChanged(nameof(BluePoints));
                OnPropertyChanged(nameof(RedPoints));
            });
        }

        private void GameModel_EntityMoved(object? sender, MoveEventArgs e)
        {
            InvokeExecuteOnMainThread(() =>
            {
                UpperFieldType entity = Fields[e.FromY * Width + e.FromX].UpperLayer;

                Fields[e.ToY * Width + e.ToX].UpperLayer = entity;
                Fields[e.ToY * Width + e.ToX].Dir = e.Direction;
                Fields[e.FromY * Width + e.FromX].UpperLayer = UpperFieldType.Empty;
            });
        }

        private void GameModel_EntityDies(object? sender, PositionEventArgs e)
        {
            InvokeExecuteOnMainThread(() =>
            {
                Fields[e.Y * Width + e.X].UpperLayer = UpperFieldType.Empty;
            });
        }

        private void GameModel_FieldChanged(object? sender, FieldChangedEventArgs e)
        {
            InvokeExecuteOnMainThread(() =>
            {
                Fields[e.Y * Width + e.X].LowerLayer = e.Type;
            });
        }

        private void GameModel_BombExplosion(object? sender, PositionEventArgs e)
        {
            InvokeExecuteOnMainThread(() =>
            {
                Fields[e.Y * Width + e.X].UpperLayer = UpperFieldType.Explosion;
            });
        }

        private void GameModel_CompetitionOver(object? sender, PlayerEventArgs e)
        {
            InvokeExecuteOnMainThread(() =>
            {
                OnPropertyChanged(nameof(BluePoints));
                OnPropertyChanged(nameof(RedPoints));
                CompetitionOver?.Invoke(this, e);
                _ = SwapToMenuPage();
            });
        }

        private void GameModel_BattleRoyalTimeChanged(object? sender, IntEventArgs e)
        {
            InvokeExecuteOnMainThread(() =>
            {
                BattleRoyalTimeLeft = e.Value;
            });
        }

        private void GameModel_PlayerStatusChanged(object? sender, PlayerStatusEventArgs e)
        {
            InvokeExecuteOnMainThread(() =>
            {
                if (e.Player == PlayerType.Blue)
                {
                    if (e.Power == SpecialPower.Invincibility)
                        BlueIsInvincible = e.Show;
                    else
                        BlueIsGhost = e.Show;
                }
                else
                {
                    if (e.Power == SpecialPower.Invincibility)
                        RedIsInvincible = e.Show;
                    else
                        RedIsGhost = e.Show;
                }
            });
        }

        #endregion

        #region Event triggers

        private void InvokeExecuteOnMainThread(Action action)
        {
            ExecuteOnMainThread?.Invoke(this, new(action));
        }

        #endregion
    }
}
